

// Const pointers

// A const before the * means "protect the thing that I point to."

// A const after the * means "protect the pointer itself."

// const at the end of the function means that you can call the function on
// const objects - that means that you can't change data members (unless
// they are mutable). This const protected "this" object.

int strcmp(const char* str, const char* otherStr);

const Car myCar;

myCar.someFunction(..);
myCar.anotherFunction(..); // illegal
 
void myFunction(Car* arg)
{
}

void Car::doNothing()
{
}

int stringLength(char* str)
{
   int count = 0;
   while(*str != 0)
   {
      ++count;
      ++str;
   }

   return count;
}

Car Car::anotherFunction(const Car* firstCar, Car* const secondCar); 

const Car& Car::someFunction(const Car* firstCar, Car* const secondCar) const
{
   Car localCar;
   Car* localPtr;

   *firstCar = localCar;  // illegal
   *secondCar = localCar; // legal

   firstCar = &localCar;  // legal
   secondCar = &localCar; // illegal

   firstCar++;   // legal
   secondCar++;   // illegal
   
   localCar = *firstCar;        // legal
   mySecondCar_ = *firstCar;    //  illegal
   
   localPtr = firstCar;     // illegal
   localPtr = const_cast<Car*>(firstCar);  // legal!

   myFunction(firstCar);   // illegal

   doNothing();  // illegal - const function cannot call nonconst function
   
   const char greeting[] = "hello";
   
   cout << stringLength( const_cast<char*>(greeting)) << endl;
   
   localPtr = secondCar;  // legal
   
   return *firstCar;  // legal (we are returning a const reference.)
   //return localCar;   // legal, but really bad - localCar will be
   // deallocated before anyone can use a reference to it.
}

class SortAscendingWithTarget
{
   public:
      SortAscendingWithTarget(int target);
      bool operator()(int first, int second);
      
   private:
      int target_;
};

SortAscendingWithTarget::SortAscendingWithTarget(int target) : target_(target)
{
}

bool SortAscendingWithTarget::operator()(int first, int second)
{
   bool returnValue = first <= second;
   
   if( first == target_) returnValue = false;
   if( second == target_) returnValue = true;
   
   return returnValue;
}

      
   

template<typename Type>
bool sortAscending(int first, int second)
{
   bool returnValue = first <= second;
   
   if( first == 100) returnValue = false;
   if( second == 100) returnValue = true;
   
   return returnValue;
}

int functorExample(vector<int>& myInts)
{
   SortAscendingWithTarget myFunctor(100);
   sort(myInts.begin(), myInts.end(), myFunctor);
   sort(myInts.begin(), myInts.end(), SortAscendingWithTarget(100));
}


// Use for_each to print every string in a container except the
// string "cs493."

void printAllButCS493(string printString)
{
   if (printString != string("cs493"))
   {
      cout << printString << endl;
   }
}

for_each(myContainer.begin(), myContainer.end(), printAllButCS493);

// More generally, print to any output stream and supress any single string.

class SuppressOne
{
   public:
      SuppressOne(ostream& stream, const string& suppressed);
      operator()(const string& printString);

   private:
      string suppressed_
      ostream& out_;
};

SuppressOne::SuppressOne(ostream& stream, const string& suppressed) :
   suppressed_(suppressed), out_(stream)
{
}

void operator()(const string& printString)
{
   if (printString != suppressed_)
   {
      out_ << printString << endl;
   }
}


for_each(myContainer.begin(), myContainer.end(), 
	 SuppressOne(cerr, string("cs493")));


// When does a copy constructor get called?
//  1.) Parameter passing by value.
//  2.) Returning function results by value.
//  3.) Initialization
//           Car myNewCar(myNeighborsCar);  // explicitly calls the copy
//                                             constructor
//           Car myNewCar = myNeighborsCar;  // implicitly calls the copy
//                                             constructor
//           Car myNewCar;   // Default constructor
//           myNewCar = myNeighborsCar; // equal operator



vector<int>  myOtherVector;
for(int i=0;i<myVector.size();i++)
{
     myOtherVector.push_back(myVector[i]  + 1);
}

int inc(int x)
{
   return ++x;
}

transform(myVector.begin(), myVector.end(), 
	  back_inserter(myOtherVector), inc);

// operator++() is preincrement - returns the value after increment.
//              ++x - where x is an instance of your class.
// operator++(int i) is postincrement - returns the value before increment.
//              x++ - where x is an instance of your class.  i will always
//              be zero.


MyClass MyClass::operator++(int i)
{
   // Store the initial value
   MyClass initialValue = *this;
   
   // Add one to this instance.
   // ....
   
   return initialValue;
}

MyClass MyClass::operator++()
{
   // Add one to this instance.
   // ....
   
   return *this;
}



